@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_process_query.c,v 1.249 2010/09/06 05:03:21 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_process_query.c,v 1.249.2.7 2011/02/23 03:56:38 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2010	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -40,6 +40,7 @@
 #include <string.h>
 #include <netinet/in.h>
 #include <ctype.h>
+#include <regex.h>
 
 #include "pool.h"
 #include "pool_config.h"
@@ -65,7 +66,7 @@
 
 static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt);
 static char *get_insert_command_table_name(InsertStmt *node);
-static int send_deallocate(POOL_CONNECTION_POOL *backend, PreparedStatementList p, int n);
+static int send_deallocate(POOL_CONNECTION_POOL *backend, POOL_SENT_MESSAGE_LIST msglist, int n);
 static int is_cache_empty(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static POOL_STATUS ParallelForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION *backend, char *database, bool send_to_frontend);
 static bool is_panic_or_fatal_error(const char *message, int major);
@@ -781,7 +782,7 @@ POOL_STATUS pool_parallel_exec(POOL_CONNECTION *frontend,
 
 
 /*
- * send SimpleQuery message to a node.
+ * send simple query message to a node.
  */
 POOL_STATUS send_simplequery_message(POOL_CONNECTION *backend, int len, char *string, int major)
 {
@@ -796,6 +797,7 @@ POOL_STATUS send_simplequery_message(POOL_CONNECTION *backend, int len, char *st
 
 	if (pool_write_and_flush(backend, string, len) < 0)
 	{
+		pool_error("send_simplequery_message: failed to send query: %s", string);
 		return POOL_END;
 	}
 
@@ -1325,8 +1327,6 @@ void reset_variables(void)
 {
 	if(pool_get_session_context())
 		pool_unset_query_in_progress();
-
-	receive_extended_begin = 0;
 }
 
 
@@ -1337,7 +1337,7 @@ void reset_variables(void)
 void reset_connection(void)
 {
 	reset_variables();
-	pool_clear_prepared_statement_list();
+	pool_clear_sent_message_list();
 }
 
 
@@ -1377,36 +1377,45 @@ static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
 	 */
 	if (qcnt >= qn)
 	{
-		if (session_context->pstmt_list.size == 0)
+		if (session_context->message_list.size == 0)
 			return 2;
 
-		char *name = session_context->pstmt_list.pstmts[0]->name;
+		char kind = session_context->message_list.sent_messages[0]->kind;
+		char *name = session_context->message_list.sent_messages[0]->name;
 
-		/* Delete from prepared list */
-		if (send_deallocate(backend, session_context->pstmt_list, 0))
+		if ((kind == 'P' || kind == 'Q') && *name != '\0')
 		{
-			/* Deallocate failed. We are in unknown state. Ask caller
-			 * to reset backend connection.
+			/* Deallocate a prepared statement */
+			if (send_deallocate(backend, session_context->message_list, 0))
+			{
+				/* Deallocate failed. We are in unknown state. Ask caller
+				 * to reset backend connection.
+				 */
+#ifdef NOT_USED
+				reset_prepared_list(&prepared_list);
+#endif
+				pool_remove_sent_message(kind, name);
+				return -1;
+			}
+			/*
+			 * If DEALLOCATE returns ERROR response, instead of
+			 * CommandComplete, del_prepared_list is not called and the
+			 * prepared object keeps on sitting on the prepared list. This
+			 * will cause infinite call to reset_backend.  So we call
+			 * del_prepared_list() again. This is harmless since trying to
+			 * remove same prepared object will be ignored.
 			 */
 #ifdef NOT_USED
-			reset_prepared_list(&prepared_list);
+			del_prepared_list(&prepared_list, prepared_list.portal_list[0]);
 #endif
-			pool_remove_prepared_statement_by_pstmt_name(name);
-			return -1;
+			pool_remove_sent_message(kind, name);
+			return 1;
+		}
+		else
+		{
+			pool_remove_sent_message(kind, name);
+			return 0;
 		}
-		/*
-		 * If DEALLOCATE returns ERROR response, instead of
-		 * CommandComplete, del_prepared_list is not called and the
-		 * prepared object keeps on sitting on the prepared list. This
-		 * will cause infinite call to reset_backend.  So we call
-		 * del_prepared_list() again. This is harmless since trying to
-		 * remove same prepared object will be ignored.
-		 */
-#ifdef NOT_USED
-		del_prepared_list(&prepared_list, prepared_list.portal_list[0]);
-#endif
-		pool_remove_prepared_statement_by_pstmt_name(name);
-		return 1;
 	}
 
 	query = pool_config->reset_query_list[qcnt];
@@ -1438,20 +1447,6 @@ static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
 	return 1;
 }
 
-/*
- * return non 0 if load balance is possible
- */
-int load_balance_enabled(POOL_CONNECTION_POOL *backend, Node* node, char *sql)
-{
-	return (pool_config->load_balance_mode &&
-			(DUAL_MODE || pool_config->parallel_mode) &&
-			MAJOR(backend) == PROTO_MAJOR_V3 &&
-			TSTATE(backend, MASTER_NODE_ID) == 'I' &&
-			is_select_query(node, sql) &&
-			!is_sequence_query(node));
-}
-
-
 /*
  * returns non 0 if the SQL statement can be load
  * balanced. Followings are statemnts go into this category.
@@ -2367,6 +2362,16 @@ POOL_STATUS do_query(POOL_CONNECTION *backend, char *query, POOL_SELECT_RESULT *
 
 		pool_debug("do_query: kind: %c", kind);
 
+		if (kind ==  'E')
+		{
+			char *message;
+
+			if (pool_extract_error_message(false, backend, major, true, &message))
+			{
+				pool_log("do_query: error message from backend %s", message);
+			}
+		}
+
 		if (major == PROTO_MAJOR_V3)
 		{
 			if (pool_read(backend, &len, sizeof(len)) < 0)
@@ -2713,25 +2718,28 @@ int need_insert_lock(POOL_CONNECTION_POOL *backend, char *query, Node *node)
 }
 
 /*
- * If a transaction has not already started, start a new one.
- * issue LOCK TABLE IN SHARE ROW EXCLUSIVE MODE if lock_kind == 1.
+ * Issue LOCK TABLE IN SHARE ROW EXCLUSIVE MODE if lock_kind == 1.
  * Issue row lock if lock_kind == 2.
  */
 POOL_STATUS insert_lock(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, char *query, InsertStmt *node, int lock_kind)
 {
 	char *table;
+	int len = 0;
 	char qbuf[1024];
 	POOL_STATUS status;
 	int i, deadlock_detected = 0;
 
-#define SEQUENCETABLEQUERY "SELECT attname FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum) WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s' AND d.adsrc ~ 'nextval'"
+#define SEQUENCETABLEQUERY "SELECT d.adsrc FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum) WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s' AND d.adsrc ~ 'nextval'"
 
-#define SEQUENCETABLEQUERY2 "SELECT attname FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum) WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s') AND d.adsrc ~ 'nextval'"
+#define SEQUENCETABLEQUERY2 "SELECT d.adsrc FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum) WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s') AND d.adsrc ~ 'nextval'"
 
 #define MAX_SEQ_NAME 128
 
-	char *atrname;
+	char *adsrc;
 	char seq_rel_name[MAX_SEQ_NAME+1];
+	regex_t preg;
+	size_t nmatch = 2;
+	regmatch_t pmatch[nmatch];
 	static POOL_RELCACHE *relcache;
 
 	/* insert_lock can be used in V3 only */
@@ -2779,14 +2787,39 @@ POOL_STATUS insert_lock(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend
 		/*
 		 * Search relcache.
 		 */
-		atrname = pool_search_relcache(relcache, backend, table);
-		if (atrname == NULL)
+		adsrc = pool_search_relcache(relcache, backend, table);
+		if (adsrc == NULL)
 		{
-			/* could not get attribute namme */
+			/* could not get adsrc */
 			return POOL_CONTINUE;
 		}
-		snprintf(seq_rel_name, MAX_SEQ_NAME, "%s_%s_seq", table, atrname);
-		pool_debug("seq rel name:%s", seq_rel_name);
+		pool_debug("adsrc: %s", adsrc);
+
+		if (regcomp(&preg, "nextval\\('(.+)'", REG_EXTENDED|REG_NEWLINE) != 0)
+		{
+			pool_error("insert_lock: regex compile failed");
+			return POOL_END;
+		}
+
+		if (regexec(&preg, adsrc, nmatch, pmatch, 0) == 0)
+		{
+			/* pmatch[0] is "nextval('...'", pmatch[1] is sequence name */
+			if (pmatch[1].rm_so >= 0 && pmatch[1].rm_eo >= 0)
+			{
+				len = pmatch[1].rm_eo - pmatch[1].rm_so;
+				strncpy(seq_rel_name, adsrc + pmatch[1].rm_so, len);
+				*(seq_rel_name + len) = '\0';
+			}
+		}
+		regfree(&preg);
+
+		if (len == 0)
+		{
+			pool_error("insert_lock: regex no match: pg_attrdef is %s", adsrc);
+			return POOL_END;
+		}
+
+		pool_debug("seq rel name: %s", seq_rel_name);
 		snprintf(qbuf, sizeof(qbuf), "SELECT 1 FROM %s FOR UPDATE", seq_rel_name);
 	}
 	else
@@ -2825,6 +2858,8 @@ POOL_STATUS insert_lock(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend
 									MASTER_CONNECTION(backend)->pid, MASTER_CONNECTION(backend)->key, 0);
 			else
 			{
+				per_node_statement_log(backend, i, qbuf);
+
 				if (lock_kind == 1)
 				{
 					status = do_command(frontend, CONNECTION(backend, i), qbuf, PROTO_MAJOR_V3, 
@@ -3250,15 +3285,20 @@ POOL_STATUS read_kind_from_backend(POOL_CONNECTION *frontend, POOL_CONNECTION_PO
 
 							if (IsA(node, DeallocateStmt))
 							{
-								PreparedStatement *ps;
+								POOL_SENT_MESSAGE *sent_msg;
 								DeallocateStmt *d = (DeallocateStmt *)node;
 
-								ps = pool_get_prepared_statement_by_pstmt_name(d->name);
-								if (ps && ps->qctxt->original_query)
+								sent_msg = pool_get_sent_message('Q', d->name);
+								if (!sent_msg)
+									sent_msg = pool_get_sent_message('P', d->name);
+								if (sent_msg)
 								{
-									string_append_char(msg, "[");
-									string_append_char(msg, ps->qctxt->original_query);
-									string_append_char(msg, "]");
+									if (sent_msg->query_context->original_query)
+									{
+										string_append_char(msg, "[");
+										string_append_char(msg, sent_msg->query_context->original_query);
+										string_append_char(msg, "]");
+									}
 								}
 							}
 						}
@@ -3299,25 +3339,26 @@ POOL_STATUS read_kind_from_backend(POOL_CONNECTION *frontend, POOL_CONNECTION_PO
 /*
  * Send DEALLOCATE message to backend by using SimpleQuery.
  */
-static int send_deallocate(POOL_CONNECTION_POOL *backend, PreparedStatementList p,
-					int n)
+static int send_deallocate(POOL_CONNECTION_POOL *backend,
+						   POOL_SENT_MESSAGE_LIST msglist, int n)
 {
-	char *query;
 	int len;
-	PrepareStmt *p_stmt;
+	char *name;
+	char *query;
 
-	if (p.size <= n)
+	if (msglist.size <= n)
 		return 1;
 
-	p_stmt = (PrepareStmt *)p.pstmts[n]->qctxt->parse_tree;
-	len = strlen(p_stmt->name) + 14; /* "DEALLOCATE \"" + "\"" + '\0' */
+	name = msglist.sent_messages[n]->name;
+
+	len = strlen(name) + 14; /* "DEALLOCATE \"" + "\"" + '\0' */
 	query = malloc(len);
 	if (query == NULL)
 	{
 		pool_error("send_deallocate: malloc failed");
 		return -1;
 	}
-	sprintf(query, "DEALLOCATE \"%s\"", p_stmt->name);
+	sprintf(query, "DEALLOCATE \"%s\"", name);
 
 	if (SimpleQuery(NULL, backend, strlen(query)+1, query) != POOL_CONTINUE)
 	{
